# Copyright 2020 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from typing_extensions import Literal


class MetadataFilter(logging.Filter):
    """Logging filter that adds constant values to each record.

    Args:
        metadata: Key-value pairs to add to each record.
    """

    def __init__(self, metadata: dict):
        self._metadata = metadata

    def filter(self, record: logging.LogRecord) -> Literal[True]:
        for key, value in self._metadata.items():
            setattr(record, key, value)
        return True
