<html lang="en">
<head>
<title>Authentication</title>
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
<link rel="stylesheet" href="/_matrix/static/client/register/style.css">
</head>
<body>
<form id="registrationForm" method="post" action="{{ myurl }}">
    <div>
        {% if error is defined %}
            <p class="error"><strong>Error: {{ error }}</strong></p>
        {% endif %}
        <p>
           Please enter a registration token.
        </p>
        <input type="hidden" name="session" value="{{ session }}" />
        <input type="text" name="token" />
        <input type="submit" value="Authenticate" />
    </div>
</form>
</body>
</html>
