<html lang="en">
<head>
    <title>Password reset failure</title>
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
</head>
<body>
<p>The request failed for the following reason: {{ failure_reason }}.</p>

<p>Your password has not been reset.</p>
</body>
</html>
