<html lang="en">
<head>
    <title>Password reset confirmation</title>
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
</head>
<body>
<!--Use a hidden form to resubmit the information necessary to reset the password-->
<form method="post">
    <input type="hidden" name="sid" value="{{ sid }}">
    <input type="hidden" name="token" value="{{ token }}">
    <input type="hidden" name="client_secret" value="{{ client_secret }}">

    <p>You have requested to <strong>reset your Matrix account password</strong>. Click the link below to confirm this action. <br /><br />
        If you did not mean to do this, please close this page and your password will not be changed.</p>
    <p><button type="submit">Confirm changing my password</button></p>
</form>
</body>
</html>

