<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Request to add an email address to your Matrix account</title>
</head>
<body>
    <p>A request to add an email address to your Matrix account has been received. If this was you, please click the link below to confirm adding this email:</p>
    <a href="{{ link }}">{{ link }}</a>
    <p>If this was not you, you can safely ignore this email. Thank you.</p>
</body>
</html>
