<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="UTF-8">
        <title>SSO account deactivated</title>
        <meta http-equiv="X-UA-Compatible" content="IE=edge">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">        <style type="text/css">
            {% include "sso.css" without context %}
        </style>
    </head>
    <body class="error_page">
        <header>
            <h1>Your account has been deactivated</h1>
            <p>
                <strong>No account found</strong>
            </p>
            <p>
                Your account might have been deactivated by the server administrator.
                You can either try to create a new account or contact the server’s
                administrator.
            </p>
        </header>
        {% include "sso_footer.html" without context %}
    </body>
</html>
