# Copyright 2022 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import Iterable

from prometheus_client import Metric

try:
    from prometheus_client.registry import Collector
except ImportError:
    # prometheus_client.Collector is new as of prometheus 0.14. We redefine it here
    # for compatibility with earlier versions.
    class _Collector(ABC):
        @abstractmethod
        def collect(self) -> Iterable[Metric]:
            pass

    Collector = _Collector  # type: ignore
