<html lang="en">
    <head>
        <title>Password reset</title>
        <meta http-equiv="X-UA-Compatible" content="IE=edge">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
    </head>
<body>
    <p>A password reset request has been received for your Matrix account. If this was you, please click the link below to confirm resetting your password:</p>

    <a href="{{ link }}">{{ link }}</a>

    <p>If this was not you, <strong>do not</strong> click the link above and instead contact your server administrator. Thank you.</p>
</body>
</html>
